@@ -1,17 +1,28 @@
+"""Adaptive Scheduler."""
+
 from adaptive_scheduler import client_support, scheduler, server_support, utils
+from adaptive_scheduler._executor import SlurmExecutor, SlurmTask
 from adaptive_scheduler._version import __version__
 from adaptive_scheduler.scheduler import PBS, SLURM
-from adaptive_scheduler.server_support import RunManager, slurm_run, start_one_by_one
+from adaptive_scheduler.server_support import (
+    MultiRunManager,
+    RunManager,
+    slurm_run,
+    start_one_by_one,
+)
 
 __all__ = [
+    "PBS",
+    "SLURM",
+    "MultiRunManager",
+    "RunManager",
+    "SlurmExecutor",
+    "SlurmTask",
     "__version__",
     "client_support",
-    "PBS",
-    "RunManager",
     "scheduler",
     "server_support",
     "slurm_run",
-    "SLURM",
     "start_one_by_one",
     "utils",
 ]
